from tc_messageBroker.rabbit_mq.event import Event
from tc_messageBroker.rabbit_mq.queue import Queue
from tc_messageBroker.rabbit_mq.status import Status
from .transaction_base import ITransaction


DISCORD_UPDATE_CHANNELS_TRANSACTIONS = [
    ITransaction(
        Queue.DISCORD_BOT,
        Event.DISCORD_BOT.FETCH,
        order=1,
        status=Status.NOT_STARTED,
    ),
    ITransaction(
        Queue.DISCORD_ANALYZER,
        Event.DISCORD_ANALYZER.RUN,
        order=2,
        status=Status.NOT_STARTED,
    ),
    # ITransaction(
    #     Queue.DISCORD_ANALYZER,
    #     Event.DISCORD_ANALYZER.SAVE,
    #     order=3,
    #     status=Status.NOT_STARTED,
    # ),
    # ITransaction(
    #     Queue.SERVER_API,
    #     Event.SERVER_API.UPDATE_GUILD,
    #     order=3,
    #     status=Status.NOT_STARTED,
    # ),
]

DISCORD_SCHEDULED_JOB_TRANSACTIONS = [
    # ITransaction(
    #     Queue.DISCORD_BOT,
    #     Event.DISCORD_BOT.FETCH,
    #     order=1,
    #     status=Status.NOT_STARTED,
    # ),
    ITransaction(
        Queue.DISCORD_ANALYZER,
        Event.DISCORD_ANALYZER.RUN_ONCE,
        order=1,
        status=Status.NOT_STARTED,
    ),
]



from .choreography import ChoreographyDict
from .saga_base import Saga
from tc_messageBroker.rabbit_mq.status import Status
from datetime import datetime


class DiscordUpdateChannel(Saga):
    def __init__(self, data: any) -> None:
        super().__init__(
            ChoreographyDict.DISCORD_UPDATE_CHANNELS,
            Status.NOT_STARTED,
            data=data,
            created_at=datetime.now(),
        )


class DiscordScheculedJob(Saga):
    def __init__(self, data: any) -> None:
        super().__init__(
            ChoreographyDict.DISCORD_SCHEDULED_JOB,
            Status.NOT_STARTED,
            data=data,
            created_at=datetime.now(),
        )



from tc_messageBroker.rabbit_mq.queue import Queue


def test_queue():
    assert Queue.SERVER_API == "SERVER_API"
    assert Queue.DISCORD_ANALYZER == "DISCORD_ANALYZER"
    assert Queue.DISCORD_BOT == "DISCORD_BOT"



from .choreography_base import IChoreography
from .transaction_base import ITransaction
from .utils.saga_base_utils import get_transactions, convert_tx_dict
from tc_messageBroker.rabbit_mq.status import Status
import uuid
from datetime import datetime
import numpy as np
from tc_messageBroker.rabbit_mq.db_operations import MongoDB
import logging


class Saga:
    def __init__(
        self,
        choreography: IChoreography,
        status: str,
        data: any,
        created_at: datetime,
        sagaId: str = None,
    ) -> None:
        self.uuid = sagaId if sagaId is not None else str(uuid.uuid1())
        self.choreography = choreography
        self.status = status
        self.data = data
        self.created_at = created_at

    def start(
        self,
        publish_method: callable,
        mongo_creds: dict[str, any],
        test_mode=False,
    ):
        """
        just publish the first transaction
        """
        tx_sorted, _ = self._sort_transactions(self.choreography.transactions)
        current_tx = tx_sorted[0]
        self.status = Status.IN_PROGRESS

        self._update_save(
            transactions=tx_sorted,
            mongo_creds=mongo_creds,
            test=test_mode,
        )

        publish_method(
            queue_name=current_tx.queue,
            event=current_tx.event,
            content={"uuid": self.uuid},
        )

    def next(
        self,
        publish_method: callable,
        call_function: callable,
        mongo_creds: dict[str, any],
        test_mode=False,
    ):
        """
        calling the next transaction within the saga
        and updating the saga state in db

        Parameters:
        ------------
        publish_method : RabbitMQ.publish | RabbitMQ.publish_on_exchange
            the publish methods that are from the RabbitMQ
        call_function : callable
            a function to be called when the message recieved
        mongo_creds : dict[str, any]
            the mongodb credentials to update the db
            the keys must be `connection_str`, `db_name`, and `collection_name`
        test_mode : bool
            testing the function indicates that we wouldn't read or write on DB
            default is False
        """
        tx_sorted, tx_not_started_count = self._sort_transactions(
            self.choreography.transactions
        )

        self.status = Status.IN_PROGRESS

        ## get the first order transaction
        current_tx = tx_sorted[0]

        current_tx.status = Status.IN_PROGRESS
        current_tx.start = datetime.now()

        self._update_save(
            transactions=tx_sorted, mongo_creds=mongo_creds, test=test_mode
        )

        try:
            ## the function we would call
            result = call_function()

            current_tx.status = Status.SUCCESS
            current_tx.end = datetime.now()
            ## in miliseconds format
            current_tx.runtime = (current_tx.end - current_tx.start).microseconds / 1000

            ## if we ran the last transaction
            if tx_not_started_count == 1:
                self.status = Status.SUCCESS
            else:
                next_tx = tx_sorted[1]
                publish_method(
                    queue_name=next_tx.queue,
                    event=next_tx.event,
                    content={"uuid": self.uuid, "data": result},
                )

            self._update_save(
                transactions=tx_sorted,
                mongo_creds=mongo_creds,
                test=test_mode,
            )

        except Exception as exp:
            logging.info(f"Error occured during the next function. Exception: {exp}")
            current_tx.error = str(exp)

            if current_tx.status != Status.SUCCESS:
                current_tx.status = Status.FAILED

            self.status = Status.FAILED

            self._update_save(
                transactions=tx_sorted,
                mongo_creds=mongo_creds,
                test=test_mode,
            )

    def _sort_transactions(self, transactions: list[ITransaction]):
        """
        sort transactions by their order and status
        the NOT_STARTED ones would be at the first of the list
        and they are ordered by `order` property

        Parameters:
        ------------
        transactions : list[ITransaction]
            the list of transactions to order

        Returns:
        ---------
        transactions_ordered : ndarray(ITransaction)
            the transactions ordered by status
            the `NOT_STARTED` ones are the firsts
            it is actually a numpy array for us to be able to
              change the properties in deep memory
        tx_not_started_count : int
            the not started transactions count
        """
        tx_not_started = []
        tx_other = []

        for tx in transactions:
            if tx.status == Status.NOT_STARTED:
                tx_not_started.append(tx)
            else:
                tx_other.append(tx)

        tx_not_started_count = len(tx_not_started)
        tx_not_started_sorted = self._sort_transactions_orderly(tx_not_started)

        transactions_ordered = list(tx_not_started_sorted)
        transactions_ordered.extend(tx_other)

        return np.array(transactions_ordered), tx_not_started_count

    def _sort_transactions_orderly(self, transactions: list[ITransaction]):
        """
        sort transactions by their `order` property

        Parameters:
        ------------
        transactions : list[ITransaction]
            the list of transactions to order

        Returns:
        ---------
        transactions_orderly_sorted : list[ITransaction]
            transactions sorted by their order
        """
        orders = [tx.order for tx in transactions]
        sorted_indices = np.argsort(orders)

        return np.array(transactions)[sorted_indices]

    def _update_save(
        self, transactions: list[ITransaction], mongo_creds: dict[str, any], test=False
    ):
        """
        update the transactions in saga choreography and then save data to db

        test is the test mode which it does not interact with db.
        default is False meaning we have to have an interaction with db
        """
        if not test:
            ## save the status into DB
            mongodb = self._get_mongo_db(
                mongo_creds=mongo_creds,
            )
            data = self._create_data()

            ## creating a duplicate choreography to avoid shallow copy
            choreography = IChoreography(
                name=self.choreography.name, transactions=transactions
            )

            self.choreography = choreography

            ## update the available choreography
            mongodb.replace(sagaId=self.uuid, data=data)

    def _create_data(self) -> dict[str, any]:
        """
        create the data from the properties of the class
          as we had in db schema

        Returns:
        ---------
        data : dict[str, any]
            the dictionary representing the current saga document

        """
        data = {}

        data["choreography"] = {}
        data["choreography"]["name"] = self.choreography.name
        data["choreography"]["transactions"] = []
        for tx in self.choreography.transactions:
            transaction = convert_tx_dict(tx)
            data["choreography"]["transactions"].append(transaction)

        data["status"] = self.status
        data["data"] = self.data
        data["sagaId"] = self.uuid
        data["createdAt"] = self.created_at
        data["updatedAt"] = datetime.now()

        return data

    def _get_mongo_db(self, mongo_creds: dict[str, any], test=False):
        """
        get mongodb instance
        """
        ## save the status into DB
        mongodb = MongoDB(
            connection_str=mongo_creds["connection_str"],
            db_name=mongo_creds["db_name"],
            collection_name=mongo_creds["collection_name"],
        )
        mongodb.connect()
        return mongodb


def get_saga(sagaId: str, connection_url: str, db_name: str, collection: str):
    """
    get saga object for a special guild

    Parameters:
    ------------
    sagaId : str
        the sagaId which the saga belongs to
    connection_url : str
        the connection to db which the saga architecture is saved
    db_name : str
        the database name to use
    collection : str
        the collection which the saga is saved

    Returns:
    ----------
    saga_obj : Saga
        the saga object to use
    """
    mongodb = MongoDB(
        connection_str=connection_url, db_name=db_name, collection_name=collection
    )
    mongodb.connect()

    data = mongodb.read(query={"sagaId": sagaId}, count=1)

    saga_obj = None
    try:
        transactions = get_transactions(data["choreography"]["transactions"])

        choreography = IChoreography(
            name=data["choreography"]["name"],
            transactions=transactions,
        )
        saga_obj = Saga(
            choreography=choreography,
            status=data["status"],
            created_at=data["createdAt"],
            sagaId=data["sagaId"],
            data=data["data"],
        )
    except Exception as exp:
        logging.error(f"Error occured! Exception: {exp}")

    return saga_obj



from tc_messageBroker.rabbit_mq.event import Event


def test_enum_event_default():
    ## For now we have the events below
    assert Event.SERVER_API.UPDATE_GUILD == "UPDATE_GUILD"
    assert Event.DISCORD_BOT.FETCH == "FETCH"
    assert Event.DISCORD_BOT.SEND_MESSAGE == "SEND_MESSAGE"
    assert Event.DISCORD_ANALYZER.RUN == "RUN"
    assert Event.DISCORD_ANALYZER.RUN_ONCE == "RUN_ONCE"
    # assert Event.DISCORD_ANALYZER.SAVE == "SAVE"



# Doing the object to dict and vice operations here
from tc_messageBroker.rabbit_mq.saga.transactions import ITransaction


def get_transactions(transactions: dict[dict[str, any]]) -> list[ITransaction]:
    transactions_obj = []

    for tx in transactions:
        transaction = ITransaction(**tx)
        transactions_obj.append(transaction)

    return transactions_obj


def convert_tx_dict(transaction: ITransaction):
    """
    convert the transaction into a dictionary
    """
    tx_dict = {}

    tx_dict["queue"] = transaction.queue
    tx_dict["event"] = transaction.event
    tx_dict["order"] = transaction.order
    tx_dict["status"] = transaction.status

    if transaction.start is not None:
        tx_dict["start"] = transaction.start
    if transaction.end is not None:
        tx_dict["end"] = transaction.end
    if transaction.message is not None:
        tx_dict["message"] = transaction.message
    if transaction.runtime is not None:
        tx_dict["runtime"] = transaction.runtime
    if transaction.error is not None:
        tx_dict["error"] = transaction.error

    return tx_dict



class ITransaction:
    def __init__(
        self, queue: str, event: str, order: int, status: str, **kwargs
    ) -> None:
        """
        the transaction happening inside the saga architecture

        Parameters:
        ------------
        queue : str
            what queue the transaction belongs to
        event : str
            the event that transaction belongs to
        order : int
            the order of the transaction in the choreography
        status : int
            the state of the transaction
        **kwargs : dict
            message : any
                the message for the transaction to hold
            start : datetime
                the start time of the transaction
                Can be `None` at the initialization
            end : datetime
                the end time of the transaction
                Can be `None` at the initialization
            runtime : float
                the runtime of the transaction
                Can be `None` at the initialization
            error : Exception
                the error the transaction could encounter
                Can be `None` at the initialization
        """
        self.message = self._get_kwarg("message", kwargs)
        self.start = self._get_kwarg("start", kwargs)
        self.end = self._get_kwarg("end", kwargs)
        self.runtime = self._get_kwarg("runtime", kwargs)
        self.error = self._get_kwarg("error", kwargs)

        self.queue = queue
        self.event = event
        self.order = order
        self.status = status

    def _get_kwarg(self, key: str, kwargs: dict[str, any], default_value=None):
        if key in kwargs.keys():
            return kwargs[key]
        else:
            return default_value



from tc_messageBroker.rabbit_mq.db_operations import MongoDB
from uuid import uuid1
from numpy.random import randint
from copy import deepcopy
from datetime import datetime, timedelta
import pytest


@pytest.mark.skip(reason="No MongoDB instance available.")
def test_insert_one():
    connection_url = "mongodb://127.0.0.1:27017/"

    mongodb = MongoDB(connection_str=connection_url)
    mongodb.connect()

    data = create_random_data()

    mongodb.write(data=[data])


@pytest.mark.skip(reason="No MongoDB instance available.")
def test_insert_multiple():
    connection_url = "mongodb://127.0.0.1:27017/"

    mongodb = MongoDB(connection_str=connection_url)
    mongodb.connect()

    documents = []
    for _ in range(5):
        data = create_random_data()
        documents.append(data)

    mongodb.write(data=documents)


@pytest.mark.skip(reason="No MongoDB instance available.")
def test_read_count_one():
    """
    read from db after writing using the previous function (above)
    """
    connection_url = "mongodb://127.0.0.1:27017/"

    mongodb = MongoDB(connection_str=connection_url)
    mongodb.connect()

    results = mongodb.read(query={}, count=1)

    assert isinstance(results, dict) is True
    assert "choreography" in results.keys()
    assert "status" in results.keys()
    assert "data" in results.keys()
    assert "sagaId" in results.keys()
    assert "createdAt" in results.keys()
    assert "updatedAt" in results.keys()


@pytest.mark.skip(reason="No MongoDB instance available.")
def test_read_count_multiple():
    """
    read from db after writing using the previous function (above)
    """
    connection_url = "mongodb://127.0.0.1:27017/"

    mongodb = MongoDB(connection_str=connection_url)
    mongodb.connect()

    data_count = 5
    results = mongodb.read(query={}, count=data_count)

    assert isinstance(results, list) is True
    assert len(results) == data_count

    for i in range(data_count):
        assert "choreography" in results[i].keys()
        assert "status" in results[i].keys()
        assert "data" in results[i].keys()
        assert "sagaId" in results[i].keys()
        assert "createdAt" in results[i].keys()
        assert "updatedAt" in results[i].keys()


@pytest.mark.skip(reason="No MongoDB instance available.")
def test_replace():
    """
    test replacing a document we just inserted
    """
    connection_url = "mongodb://127.0.0.1:27017/"

    mongodb = MongoDB(connection_str=connection_url)
    mongodb.connect()

    data = create_random_data()

    mongodb.write(data=[data])

    data.pop("_id")

    new_data = deepcopy(data)

    choreography_name = "NEW_TEST"
    tx_state = "SUCCESS"

    new_data["choreography"]["name"] = choreography_name
    new_data["choreography"]["name"] = choreography_name
    new_data["choreography"]["transactions"]["0"]["status"] = tx_state

    mongodb.replace(data["sagaId"], new_data)

    document = mongodb.read({"sagaId": new_data["sagaId"]})
    document.pop("_id")

    assert document != data
    assert document == new_data


@pytest.mark.skip(reason="No MongoDB instance available.")
def test_delete():
    """
    delete all except one of the ligitimate data which we choose
    """

    connection_url = "mongodb://127.0.0.1:27017/"

    mongodb = MongoDB(connection_str=connection_url)
    mongodb.connect()

    mongodb.delete(query={"data.guildId": {"$ne": "993163081939165234"}})


@pytest.mark.skip(reason="No MongoDB instance available.")
def test_wrong_input():
    """
    give wrong input to insert
    """
    connection_url = "mongodb://127.0.0.1:27017/"

    mongodb = MongoDB(connection_str=connection_url)
    mongodb.connect()

    data = create_random_data()
    data.pop("sagaId")

    valid = mongodb._validator(data=data)
    assert valid is False


@pytest.mark.skip(reason="No MongoDB instance available.")
def create_random_data():
    random_guildId = ""
    for _ in range(18):
        random_number = randint(0, 10)
        random_guildId += str(random_number)

    data = {
        "choreography": {
            "name": f"test_{randint(0, 50)}",
            "transactions": {
                "0": {
                    "queue": "SERVER_API",
                    "event": "UPDATE_GUILD",
                    "order": 1,
                    "status": "NOT_STARTED",
                },
                "1": {
                    "queue": "DISCORD_BOT",
                    "event": "SEND_MESSAGE",
                    "order": 2,
                    "status": "NOT_STARTED",
                },
            },
        },
        "status": "NOT_STARTED",
        "data": {"guildId": random_guildId},
        "sagaId": str(uuid1()),
        "createdAt": datetime.now().isoformat(),
        "updatedAt": (datetime.now() + timedelta(minutes=2)).isoformat(),
    }

    return data



from tc_messageBroker.rabbit_mq.saga.transaction_base import ITransaction
from tc_messageBroker.rabbit_mq.queue import Queue
from tc_messageBroker.rabbit_mq.event import Event
from tc_messageBroker.rabbit_mq.status import Status
from datetime import datetime, timedelta


def test_transaction_default():
    transaction = ITransaction(
        queue=Queue.DISCORD_ANALYZER,
        event=Event.DISCORD_ANALYZER.RUN,
        order=1,
        status=Status.NOT_STARTED,
    )

    assert transaction.queue == Queue.DISCORD_ANALYZER
    assert transaction.event == Event.DISCORD_ANALYZER.RUN
    assert transaction.order == 1
    assert transaction.status == Status.NOT_STARTED
    assert transaction.message is None
    assert transaction.start is None
    assert transaction.end is None
    assert transaction.runtime is None
    assert transaction.error is None


def test_transaction_all_inputs():
    """
    transaction with all inputs given
    """
    now = datetime.now()

    transaction = ITransaction(
        queue=Queue.DISCORD_ANALYZER,
        event=Event.DISCORD_ANALYZER.RUN,
        order=1,
        status=Status.NOT_STARTED,
        message="sample_message",
        start=now - timedelta(minutes=5),
        end=now,
        runtime=10000,
        error=None,
    )

    assert transaction.queue == Queue.DISCORD_ANALYZER
    assert transaction.event == Event.DISCORD_ANALYZER.RUN
    assert transaction.order == 1
    assert transaction.status == Status.NOT_STARTED
    assert transaction.message == "sample_message"
    assert transaction.start == now - timedelta(minutes=5)
    assert transaction.end == now
    assert transaction.runtime == 10000
    assert transaction.error is None



from tc_messageBroker.rabbit_mq.saga.saga_base import Saga
from tc_messageBroker.rabbit_mq.saga.transaction_base import ITransaction
from tc_messageBroker.rabbit_mq.saga.choreography_base import IChoreography
from datetime import datetime
from tc_messageBroker.rabbit_mq.status import Status


def test_start_tx():
    choreography = IChoreography(
        name="choreography_with_random_tx",
        transactions=[
            ITransaction("queue", "event", 5, Status.NOT_STARTED),
            ITransaction("queue1", "event1", 2, Status.NOT_STARTED),
            ITransaction("queue", "event", 3, Status.SUCCESS),
            ITransaction("queue", "event", 1, Status.FAILED),
            ITransaction("queue", "event", 0, Status.CANCELLED),
        ],
    )

    saga = Saga(choreography, Status.NOT_STARTED, data=None, created_at=datetime.now())

    def sample_publish(**kwargs):
        assert kwargs["queue_name"] == choreography.transactions[1].queue
        assert kwargs["event"] == choreography.transactions[1].event
        assert kwargs["content"]["uuid"] == saga.uuid

    saga.start(
        publish_method=sample_publish,
        mongo_creds={},
        test_mode=True,
    )

    assert saga.status == Status.IN_PROGRESS
    assert saga.choreography.transactions[0].status == Status.NOT_STARTED
    assert saga.choreography.transactions[1].status == Status.NOT_STARTED
    assert saga.choreography.transactions[2].status == Status.SUCCESS
    assert saga.choreography.transactions[3].status == Status.FAILED
    assert saga.choreography.transactions[4].status == Status.CANCELLED



"""
test the saga_base module
"""
from tc_messageBroker.rabbit_mq.saga.saga_base import Saga
from tc_messageBroker.rabbit_mq.saga.choreography import ChoreographyDict
from tc_messageBroker.rabbit_mq.saga.choreography_base import IChoreography
from tc_messageBroker.rabbit_mq.saga.transaction_base import ITransaction
from datetime import datetime
from tc_messageBroker.rabbit_mq.status import Status


def test_choreagraphy_sorting_orders_predefined_choreogprahy():
    saga = Saga(
        ChoreographyDict.DISCORD_UPDATE_CHANNELS,
        Status.NOT_STARTED,
        data=None,
        created_at=datetime.now(),
    )

    ## it should sort the NOT_STARTED transactions
    (tx_sorted, not_started_count) = saga._sort_transactions(
        saga.choreography.transactions
    )

    assert not_started_count == 2
    order_val = 0
    for tx in tx_sorted:
        if tx.status == Status.NOT_STARTED:
            assert order_val < tx.order
            order_val = tx.order


def test_assert_sorting_status_predefined_choreagprahy():
    """assert the NOT_STARTED transactions to be at the top of the list"""

    saga = Saga(
        ChoreographyDict.DISCORD_UPDATE_CHANNELS,
        Status.NOT_STARTED,
        data=None,
        created_at=datetime.now(),
    )

    ## it should sort the NOT_STARTED transactions
    (tx_sorted, not_started_count) = saga._sort_transactions(
        saga.choreography.transactions
    )

    assert not_started_count == 2

    condition = False
    for tx in tx_sorted:
        if tx.status == Status.NOT_STARTED:
            assert condition is False
        else:
            condition = True


def test_choreagraphy_sorting_orders_random_choreogprahy():
    choreography = IChoreography(
        name="choreography_with_random_tx",
        transactions=[
            ITransaction("queue", "event", 5, Status.IN_PROGRESS),
            ITransaction("queue", "event", 2, Status.NOT_STARTED),
            ITransaction("queue", "event", 4, Status.NOT_STARTED),
            ITransaction("queue", "event", 3, Status.SUCCESS),
            ITransaction("queue", "event", 1, Status.FAILED),
            ITransaction("queue", "event", 0, Status.CANCELLED),
        ],
    )

    saga = Saga(choreography, Status.NOT_STARTED, data=None, created_at=datetime.now())

    ## it should sort the NOT_STARTED transactions
    (tx_sorted, not_started_count) = saga._sort_transactions(
        saga.choreography.transactions
    )

    assert not_started_count == 2
    order_val = 0
    for tx in tx_sorted:
        if tx.status == Status.NOT_STARTED:
            assert order_val < tx.order
            order_val = tx.order


def test_choreagraphy_sorting_status_random_choreogprahy():
    """assert the NOT_STARTED transactions to be at the top of the list"""

    choreography = IChoreography(
        name="choreography_with_random_tx",
        transactions=[
            ITransaction("queue", "event", 5, Status.IN_PROGRESS),
            ITransaction("queue", "event", 2, Status.NOT_STARTED),
            ITransaction("queue", "event", 4, Status.NOT_STARTED),
            ITransaction("queue", "event", 3, Status.SUCCESS),
            ITransaction("queue", "event", 1, Status.FAILED),
            ITransaction("queue", "event", 0, Status.CANCELLED),
        ],
    )

    saga = Saga(choreography, Status.NOT_STARTED, data=None, created_at=datetime.now())

    ## it should sort the NOT_STARTED transactions
    (tx_sorted, not_started_count) = saga._sort_transactions(
        saga.choreography.transactions
    )

    assert not_started_count == 2

    condition = False
    status = []
    for tx in tx_sorted:
        status.append(tx.status)
        if tx.status == Status.NOT_STARTED:
            assert condition is False
        else:
            condition = True



from .events_microservice import ServerEvent, DiscordBotEvent, DiscordAnalyzerEvent


class Event:
    SERVER_API = ServerEvent
    DISCORD_BOT = DiscordBotEvent
    DISCORD_ANALYZER = DiscordAnalyzerEvent



from tc_messageBroker.rabbit_mq.saga.transaction_base import ITransaction


class IChoreography:
    def __init__(self, name: str, transactions: list[ITransaction]) -> None:
        self.name = name
        self.transactions = transactions



from tc_messageBroker import RabbitMQ
from tc_messageBroker.rabbit_mq.queue import Queue
from tc_messageBroker.rabbit_mq.event import Event
import pytest


def test_tc_message_broker_default_mode():
    url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=url, port=port, username=username, password=password
    )

    assert messageBroker.broker_url == url


def test_tc_message_broker_empty_url():
    url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=url, port=port, username=username, password=password
    )

    assert messageBroker.broker_url == url


def test_tc_message_broker_singleton():
    url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=url, port=port, username=username, password=password
    )

    assert messageBroker.broker_url == url

    url_new = "127.0.0.1"
    messageBroker2 = RabbitMQ(
        broker_url=url_new, port=port, username=username, password=password
    )
    assert messageBroker == messageBroker2
    assert messageBroker.broker_url == url_new
    assert messageBroker2.broker_url == url_new


def test_tc_message_broker_connection_exp():
    """test the message broker with wrong url"""
    url = "wrong_url"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=url, port=port, username=username, password=password
    )
    connection_status = messageBroker.connect(queue_name="test_queue")
    assert connection_status is False


def test_tc_message_broker_set_on_event():
    """
    test the adding on event function
    """

    callback_func = lambda: 2 * 3

    url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=url, port=port, username=username, password=password
    )

    _ = messageBroker.connect(queue_name="test_queue")
    messageBroker.on_event(event_name="some_event", on_message=callback_func)


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (RabbitMQ instance not available)"
)
def test_connection_rabbit_mq():
    broker_url = "localhost"

    rabbit_mq = RabbitMQ(broker_url=broker_url)
    success = rabbit_mq.connect("test_queue")
    rabbit_mq.connection.close()

    assert success is True


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (RabbitMQ instance not available)"
)
def test_consume_publish_no_event():
    """
    consume an event which does not exists
    """
    url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    rabbit_mq = RabbitMQ(
        broker_url=url, port=port, username=username, password=password
    )
    rabbit_mq.connect(Queue.DISCORD_ANALYZER)

    content = {"content": "content_analyzer_publish"}

    rabbit_mq.publish(
        queue_name=Queue.DISCORD_ANALYZER,
        event=Event.DISCORD_ANALYZER.RUN,
        content=content,
    )
    rabbit_mq.consume(Queue.DISCORD_ANALYZER)
    rabbit_mq.connection.close()


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (start_consuming woudn't stop and test wouldn't end)"  # flake8: noqa
)
def test_consume_publish_with_event():
    """
    consume an event which does not exists
    """
    url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    rabbit_mq = RabbitMQ(
        broker_url=url, port=port, username=username, password=password
    )
    rabbit_mq.connect(Queue.DISCORD_ANALYZER)
    rabbit_mq.connect(Queue.DISCORD_ANALYZER)

    content = {"content": "content_analyzer_publish"}

    global_var = None

    def event_function():
        global global_var
        global rabbit_mq
        global_var = True

        # rabbit_mq.channel.stop_consuming(consumer_tag='SAMPLE')

    rabbit_mq.publish(
        Queue.DISCORD_ANALYZER, Event.DISCORD_ANALYZER.RUN, content=content
    )
    rabbit_mq.consume(
        Queue.DISCORD_ANALYZER, consume_options={"consumer_tag": "SAMPLE"}
    )
    rabbit_mq.channel.basic_consume(Queue.DISCORD_ANALYZER, rabbit_mq._consume_callback)

    rabbit_mq.channel.start_consuming()

    rabbit_mq.connection.close()

    assert global_var is True



import pika

# from typing import Callable
import logging
from datetime import datetime
import json
import functools


class RabbitMQ:
    def __init__(self, broker_url: str, port: int, username: str, password: str):
        logging.basicConfig()
        logging.getLogger().setLevel(logging.INFO)

        self.broker_url = broker_url
        self.port = port
        self._username = username
        self._password = password

        ## it will use the default exchange point if not created
        self.exchange_name = ""
        self.channel = None
        self.connection = None
        self.event_function = {}

    def __new__(cls, broker_url: str, port: int, username: str, password: str):
        ## making it singleton
        if not hasattr(cls, "instance"):
            cls.instance = super(RabbitMQ, cls).__new__(cls)
        return cls.instance

    def connect(
            self, 
            queue_name: str, 
            consume_options: dict = None,
            heartbeat: int = 60
        ) -> bool:
        """
        connect the rabbitMQ broker and start consuming

        Parameters:
        -------------
        queue_name : str
            the queue name to connect
        consume_options : dict
            additional arguments for basic_consume method
            default is `None`
        heartbeat : int
            the number of seconds for a message to stay alive
            default is 60 seconds

        Returns:
        ---------
        is_successful : bool
            if True, connecting to rabbitMQ server was successful
            otherwise would return False
        """
        try:
            credentials = pika.PlainCredentials(self._username, self._password)
            amqpServer = self.broker_url
            self.connection = pika.BlockingConnection(
                pika.ConnectionParameters(
                    host=amqpServer,
                    port=self.port,
                    credentials=credentials,
                    heartbeat=heartbeat, ## disabling the heartbeat
                ),
            )
            self.channel = self.connection.channel()

            # make sure that the channel is created,
            # if not this statement will create it
            self.channel.queue_declare(queue=queue_name)

            self.channel.basic_consume(
                queue=queue_name,
                on_message_callback=self._consume_callback,
                auto_ack=False,
                arguments=consume_options,
            )
            return True
        except Exception as exp:
            logging.error(f" Something went wrong with RabbitMQ {exp}")
            return False

    def _consume_callback(self, ch, method, properties, body) -> bool:
        """
        consume a message with a specific body

        Parameters:
        -------------
        method : any
            the method which have the property of delivery_tag
        body : dict[str, any]
            the dictionary containing the body of message to be consumed

        Returns:
        ---------
        is_successful : bool
            if True, the event function was called
            otherwise would return False and requeue the message
        """
        body_serialized = json.loads(body)
        event = body_serialized["event"]

        if event not in self.event_function.keys():
            logging.info(" An Event was received that doesn't exist")
            self.connection.add_callback_threadsafe(
                functools.partial(
                    self.channel.basic_reject,
                    delivery_tag=method.delivery_tag,
                    requeue=True,
                )
            )
        else:
            self.event_function[event](body_serialized)
            self.connection.add_callback_threadsafe(
                functools.partial(
                    self.channel.basic_ack,
                    delivery_tag=method.delivery_tag,
                )
            )

    def consume(self, queue_name: str, consume_options: dict = None):
        """
        set consuming events from a queue
        """
        self.channel.queue_declare(queue=queue_name)

        self.channel.basic_consume(
            queue=queue_name,
            on_message_callback=self._consume_callback,
            arguments=consume_options,
            auto_ack=False,
        )

    def publish(
        self, queue_name: str, event: str, content: dict, options: any = None
    ) -> None:
        """
        Publish a specific message to a specific queue directly

        Parameters:
        -------------
        queue_name : str
            the exchange point name
        event : str
            the event name to use
            using the events available under Event class is recommended
        content : dict
            dictionary of contents to publish
        options : any
            additional arguments for basic_publish method
            default is `None`

        """
        data = self._define_data(event=event, content=content)

        self.channel.queue_declare(queue=queue_name)

        self.channel.basic_publish(
            exchange=self.exchange_name,
            routing_key=queue_name,
            body=data,
            properties=options,
        )

    def on_event(self, event_name: str, on_message: callable) -> None:
        """
        set a function to be called when an event happened

        Parameters:
        -------------
        event_name : str
            the event name we're setting the function for
        on_message : callable
            the message when recieved (consumed) to call the function

        """
        self.event_function[event_name] = on_message

    def create_exchange(
        self,
        name: str,
        type: str,
        **kwargs,
    ) -> None:
        """
        create an exchange point with a specific type

        Parameters:
        -------------
        name : str
            the name of the exchange point
        type : str
            the type of exchange point
            Must be either one of the
                'direct', 'topic', 'headers', or 'fanout'
        **kwargs : dict
            durable : bool
                survive the exchange point if the server restarts
                if True, will be survvived
                if False won't be re-created after server restart
                default is True
            auto_delete : bool
                whether to delete the exchange point when no queue is bound to it
                if True, will delete the exchange point when no queue is bound to it
                else, will do otherwise
            options : any
                more options for the exchange_declare
        """
        ## default values
        durable = True
        auto_delete = False
        options = None

        if "durable" in kwargs.keys():
            durable = kwargs["durable"]
        if "auto_delete" in kwargs.keys():
            auto_delete = kwargs["auto_delete"]
        if "options" in kwargs.keys():
            options = kwargs["options"]

        self.channel.exchange_declare(
            exchange=name,
            exchange_type=type,
            durable=durable,
            auto_delete=auto_delete,
            arguments=options,
        )
        self.exchange_name = name

    def bind_queue_to_exchange(
        self, queue_name: str, exchange_name: str, pattern: str
    ) -> None:
        """
        bind a queue to a special exchange point

        Parameters:
        -----------
        queue_name : str
            the name of the queue we want to bind
        exchange_name : str
            the exchange point name
        pattern : str
            routing_key of the queue
        """
        self.channel.queue_declare(queue=queue_name)

        self.channel.queue_bind(
            queue=queue_name, exchange=exchange_name, routing_key=pattern
        )

    def publish_on_exchange(
        self,
        exchange_name: str,
        routing_key: str,
        event: str,
        content: dict,
        options=None,
    ) -> None:
        """
        Publish a specific message to an exchange point using a routing key

        Parameters:
        -------------
        exchange_name : str
            the exchange point name
        routing_key : str
            the routing key used to route the messages to one or more queues
        event : str
            the event name to use
            using the events available under Event class is recommended
        content : dict
            dictionary of contents to publish
        options : any
            the properties of basic_publish method

        """

        data = self._define_data(event=event, content=content)

        self.channel.basic_publish(
            exchange=exchange_name,
            routing_key=routing_key,
            body=data,
            properties=options,
        )

    def _define_data(self, event: str, content: dict) -> str:
        """
        define seriazable data to use

        Parameters:
        ------------
        event : str
            the event name to use
            using the events available under Event class is recommended
        content : dict
            dictionary of contents to publish

        Retuns:
        ---------
        data_json : str
            the data in json format wrapped into a string
        """
        data = {"event": event, "date": str(datetime.now()), "content": content}

        data_json = json.dumps(data)
        return data_json



from tc_messageBroker.rabbit_mq.saga.choreography_base import IChoreography
from tc_messageBroker.rabbit_mq.saga.transactions import (
    DISCORD_UPDATE_CHANNELS_TRANSACTIONS,
    DISCORD_SCHEDULED_JOB_TRANSACTIONS,
)


def test_choreography_discord_update_channels_no_inputs():
    choreography = IChoreography(name=None, transactions=None)

    assert choreography.name is None
    assert choreography.transactions is None


def test_choreography_discord_update_channels():
    choreography = IChoreography(
        name="sample", transactions=DISCORD_UPDATE_CHANNELS_TRANSACTIONS
    )

    assert choreography.name == "sample"
    assert choreography.transactions == DISCORD_UPDATE_CHANNELS_TRANSACTIONS


def test_choreography_discord_update_channels_wrong_input():
    choreography = IChoreography(
        name="sample", transactions=DISCORD_UPDATE_CHANNELS_TRANSACTIONS
    )

    assert choreography.name != "some_choreography"
    assert choreography.transactions != DISCORD_SCHEDULED_JOB_TRANSACTIONS


def test_choreography_discord_job_transactions():
    choreography = IChoreography(
        name="sample", transactions=DISCORD_SCHEDULED_JOB_TRANSACTIONS
    )

    assert choreography.name == "sample"
    assert choreography.transactions == DISCORD_SCHEDULED_JOB_TRANSACTIONS



from tc_messageBroker.rabbit_mq.status import Status


def test_enum_stats():
    assert Status.NOT_STARTED == "NOT_STARTED"
    assert Status.IN_PROGRESS == "IN_PROGRESS"
    assert Status.SUCCESS == "SUCCESS"
    assert Status.FAILED == "FAILED"
    assert Status.CANCELLED == "CANCELLED"



from tc_messageBroker import RabbitMQ
import pytest


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no RabbitMQ instance available)"
)
def test_tc_message_broker_declare_default_exchange_direct():
    """
    declare direct exchange point with default parameters,
     durable = True,
     auto_delete = False,
     options = None
    """

    broker_url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=broker_url, port=port, username=username, password=password
    )

    _ = messageBroker.connect(queue_name="test_queue")

    messageBroker.create_exchange(name="exchange_point", type="direct")


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no RabbitMQ instance available)"
)
def test_tc_message_broker_declare_default_exchange_topic():
    """
    declare topic exchange point with default parameters,
     durable = True,
     auto_delete = False,
     options = None
    """
    broker_url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=broker_url, port=port, username=username, password=password
    )

    _ = messageBroker.connect(queue_name="test_queue")

    messageBroker.create_exchange(name="exchange_point2", type="topic")


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no RabbitMQ instance available)"
)
def test_tc_message_broker_declare_exchange_fanout():
    """
    declare fanout exchange point with default parameters,
     durable = True,
     auto_delete = False,
     options = None
    """
    broker_url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=broker_url, port=port, username=username, password=password
    )
    _ = messageBroker.connect(queue_name="test_queue")

    messageBroker.create_exchange(name="exchange_point3", type="fanout")


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no RabbitMQ instance available)"
)
def test_tc_message_broker_declare_default_exchange_headers():
    """
    declare `match` exchange with default parameters,
     durable = True,
     auto_delete = False,
     options = None
    """
    broker_url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=broker_url, port=port, username=username, password=password
    )
    _ = messageBroker.connect(queue_name="test_queue")

    messageBroker.create_exchange(name="exchange_point4", type="headers")


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no RabbitMQ instance available)"
)
def test_tc_message_broker_declare_durable_false_exchange_direct():
    """
    declare `match` exchange with default parameters,
     durable = False,
     auto_delete = False,
     options = None
    """
    broker_url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=broker_url, port=port, username=username, password=password
    )
    _ = messageBroker.connect(queue_name="test_queue")

    messageBroker.create_exchange(
        name="exchange_point5",
        type="direct",
        durable=False,
    )


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no RabbitMQ instance available)"
)
def test_tc_message_broker_declare_auto_delete_true_exchange_direct():
    """
    declare `match` exchange with default parameters,
     durable = False,
     auto_delete = False,
     options = None
    """
    broker_url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=broker_url, port=port, username=username, password=password
    )
    _ = messageBroker.connect(queue_name="test_queue")

    messageBroker.create_exchange(
        name="exchange_point6",
        type="direct",
        auto_delete=True,
    )


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no RabbitMQ instance available)"
)
def test_tc_message_broker_publish_on_exchange():
    """
    declare direct exchange point with default parameters,
     durable = True,
     auto_delete = False,
     options = None
    """
    broker_url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=broker_url, port=port, username=username, password=password
    )
    _ = messageBroker.connect(queue_name="test_queue")

    messageBroker.create_exchange(name="exchange_point7", type="direct")
    messageBroker.publish_on_exchange(
        exchange_name="exchange_point7",
        routing_key="sample",
        event="test_event",
        content={"guildId": 123},
    )


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no RabbitMQ instance available)"
)
def test_tc_message_broker_bind_queue_on_exchange():
    """
    declare direct exchange point with default parameters,
     durable = True,
     auto_delete = False,
     options = None
    """
    broker_url = "localhost"
    port = 5672
    username = "guest"
    password = "guest"

    messageBroker = RabbitMQ(
        broker_url=broker_url, port=port, username=username, password=password
    )
    _ = messageBroker.connect(queue_name="test_queue")

    messageBroker.create_exchange(name="exchange_point8", type="direct")

    messageBroker.bind_queue_to_exchange(
        queue_name="sample_queue", exchange_name="exchange_point8", pattern="sample"
    )



from tc_messageBroker.rabbit_mq.saga.utils.saga_base_utils import (
    convert_tx_dict,
    get_transactions,
)
from tc_messageBroker.rabbit_mq.saga.transaction_base import ITransaction
from datetime import datetime, timedelta


def test_convert_transaction_to_dict_keys():
    queue = "sample_queue"
    event = "sample_event"
    order = 1
    status = "SAMPLE_STATUS"
    tx = ITransaction(
        queue=queue,
        event=event,
        order=order,
        status=status,
    )

    tx_dict = convert_tx_dict(tx)

    keys = tx_dict.keys()

    assert "queue" in keys
    assert "event" in keys
    assert "order" in keys
    assert "status" in keys


def test_convert_transaction_to_dict_values():
    queue = "sample_queue"
    event = "sample_event"
    order = 1
    status = "SAMPLE_STATUS"
    tx = ITransaction(
        queue=queue,
        event=event,
        order=order,
        status=status,
    )

    tx_dict = convert_tx_dict(tx)

    assert tx_dict["queue"] == queue
    assert tx_dict["event"] == event
    assert tx_dict["order"] == order
    assert tx_dict["status"] == status


def test_convert_transaction_to_dict_additional_params():
    queue = "sample_queue"
    event = "sample_event"
    order = 1
    status = "SAMPLE_STATUS"
    start = datetime.now() - timedelta(minutes=1)
    end = datetime.now()
    runtime = (end - start).microseconds * 1000
    message = {"data"}
    error = "Error"

    tx = ITransaction(
        queue=queue,
        event=event,
        order=order,
        status=status,
        start=start,
        end=end,
        runtime=runtime,
        message=message,
        error=error,
    )

    tx_dict = convert_tx_dict(tx)

    assert tx_dict["queue"] == queue
    assert tx_dict["event"] == event
    assert tx_dict["order"] == order
    assert tx_dict["status"] == status
    assert tx_dict["start"] == start
    assert tx_dict["end"] == end
    assert tx_dict["runtime"] == runtime
    assert tx_dict["message"] == message


def test_get_transactions_length():
    event = "EVENT"
    queue = "QUEUE"
    order = 1
    status = "STATUS"

    transactions = []
    loop_iter = 2
    for i in range(loop_iter):
        tx = {}
        tx["event"] = event + str(i)
        tx["queue"] = queue + str(i)
        tx["order"] = order + i
        tx["status"] = status + str(i)

        transactions.append(tx)

    transactions_obj = get_transactions(transactions)

    assert len(transactions_obj) == loop_iter


def test_get_transactions_obj():
    event = "EVENT"
    queue = "QUEUE"
    order = 1
    status = "STATUS"

    transactions = []
    loop_iter = 2
    for i in range(loop_iter):
        tx = {}
        tx["event"] = event + str(i)
        tx["queue"] = queue + str(i)
        tx["order"] = order + i
        tx["status"] = status + str(i)

        transactions.append(tx)

    transactions_obj = get_transactions(transactions)

    for i in range(loop_iter):
        assert isinstance(transactions_obj[i], ITransaction) is True



from .choreography_base import IChoreography
from .transactions import (
    DISCORD_UPDATE_CHANNELS_TRANSACTIONS,
    DISCORD_SCHEDULED_JOB_TRANSACTIONS,
)


DISCORD_UPDATE_CHANNELS = IChoreography(
    name="DISCORD_UPDATE_CHANNELS",
    transactions=DISCORD_UPDATE_CHANNELS_TRANSACTIONS,
)

DISCORD_SCHEDULED_JOB = IChoreography(
    name="DISCORD_UPDATE_CHANNELS",
    transactions=DISCORD_SCHEDULED_JOB_TRANSACTIONS,
)


class ChoreographyDict:
    DISCORD_UPDATE_CHANNELS = DISCORD_UPDATE_CHANNELS
    DISCORD_SCHEDULED_JOB = DISCORD_SCHEDULED_JOB



class BotBaseEvent:
    FETCH = "FETCH"


class AnalyzerBaseEvent:
    RUN = "RUN"
    SAVE = "SAVE"
    RUN_ONCE = "RUN_ONCE"


class ServerEvent:
    UPDATE_GUILD = "UPDATE_GUILD"


class DiscordBotEvent:
    FETCH = BotBaseEvent.FETCH
    SEND_MESSAGE = "SEND_MESSAGE"


class DiscordAnalyzerEvent:
    RUN = AnalyzerBaseEvent.RUN  ## RECOMPUTE
    RUN_ONCE = AnalyzerBaseEvent.RUN_ONCE
    # SAVE = AnalyzerBaseEvent.SAVE



from setuptools import setup, find_packages

with open("requirements.txt") as f:
    requirements = f.read().splitlines()


setup(
    name="tc-messageBroker",
    version="1.2.0",
    author="Mohammad Amin Dadgar, RnDAO",
    maintainer="Mohammad Amin Dadgar",
    maintainer_email="dadgaramin96@gmail.com",
    packages=find_packages(),
    description="Shared library for message broker in Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=requirements,
)



from pymongo import MongoClient
from pymongo.read_concern import ReadConcern
from pymongo.write_concern import WriteConcern
from tc_messageBroker.rabbit_mq.saga.utils.choreography_schema import (
    choreography_schema,
)
import jsonschema
from jsonschema.exceptions import ValidationError
import logging


class MongoDB:
    def __init__(self, connection_str: str, db_name: str, collection_name: str) -> None:
        """
        initialize the mongodb class

        Parameters:
        -------------
        connection_str : str
            the url to connect to mongo_db
        db_name : str
            the database we need
        collection_name : str
            name of the collection that we would use
        """
        self.connection_str = connection_str
        self.db_name = db_name
        self.collection_name = collection_name
        self.client = None

    def connect(self) -> None:
        """
        connect to MongoDB database
        """
        client = MongoClient(host=self.connection_str)
        self.client = client

    def replace(self, sagaId: str, data: dict):
        """
        replace a document with the new data

        Parameters:
        -------------
        sagaId : str
            used to find the document we're going to replace
        data : dict
            the data we're going to replace the document with
        """
        valid = self._validator([data])

        def callback_wrapper(session):
            self._write_db(
                session, data, mode="replace", replace_query={"sagaId": sagaId}
            )

        if valid:
            with self.client.start_session() as session:
                session.with_transaction(
                    callback=callback_wrapper,
                    read_concern=ReadConcern("local"),
                    write_concern=WriteConcern("local"),
                )

    def write(self, data: list[dict]):
        """
        write the data into db
        """
        valid = self._validator(data)

        if valid:
            with self.client.start_session() as session:
                session.with_transaction(
                    callback=lambda session: self._write_db(session, data),
                    read_concern=ReadConcern("local"),
                    write_concern=WriteConcern("local"),
                )

    def read(self, query: dict = {}, count: int = 1) -> list:
        """
        read from database (using find method)

        Parameters:
        -------------
        query : dict
            the query to find a specific data
            default is `{}` which would mean to return all data
        count : int
            the count of samples to get
            if 1 then findOne method is used
            else the find method in mongoClient is used

        Returns:
        ---------
        data : list | dict
            a list of data or one dictionary would be returned
            based on the count of data requested
            if count equal to 1 then one python dictionary would be returned
            else a list is going to be returned
        """
        if count == 1:
            cursor = self.client[self.db_name][self.collection_name].find_one(query)
            data = cursor
        else:
            cursor = (
                self.client[self.db_name][self.collection_name].find(query).limit(count)
            )
            data = list(cursor)

        return data

    def delete(self, query: dict):
        """
        delete documents

        Parameters:
        -------------
        query : dict
            the query to delete documents we want
        """
        with self.client.start_session() as session:
            session.with_transaction(
                callback=lambda session: self._write_db(
                    session, data=None, mode="delete", delete_query=query
                ),
                read_concern=ReadConcern("local"),
                write_concern=WriteConcern("local"),
            )

    def _write_db(self, session, data: dict, mode="add", **kwargs) -> None:
        """
        writing the data into db within a session
        (creating the transaction)

        Parameters:
        -------------
        session : MongoClient.session
            the session we want to use for writing operations on db
        data : dict
            the data we're going to write
        mode: str
            we're going to either `add`, `replace` or `delete`
        **kwargs:
            replace_query : dict
                if our mode is replace, then we should use a dictionary to query
                the document we want to replace
            delete_query : dict
                if our mode is delete, then we should use a dictionary to query
                the documents we want to delete
        """

        try:
            match mode:
                case "add":
                    session.client[self.db_name][self.collection_name].insert_many(data)
                case "replace":
                    session.client[self.db_name][self.collection_name].replace_one(
                        kwargs["replace_query"], data
                    )
                case "delete":
                    session.client[self.db_name][self.collection_name].delete_many(
                        kwargs["delete_query"]
                    )
                case _:
                    logging.error(f"The writing mode: {mode} is not implemented!")
        except KeyError as exp:
            logging.error(f"{exp}: unsufficient function inputs!")
        except Exception as exp:
            logging.error(f"Error occured: {exp}")

    def _validator(self, data: list[dict]):
        """
        validate the schema of data before saving it to db

        Parameters:
        ------------
        data : dict
            a nested python dict
        """
        try:
            for document in data:
                jsonschema.validate(
                    instance=document,
                    schema=choreography_schema,
                )
            return True
        except ValidationError as error:
            logging.error(f"Not valid schema to save in DB! Error: {error}")
            return False



from tc_messageBroker.rabbit_mq.status import Status

choreography_schema = {
    "type": "object",
    "properties": {
        "sagaId": {"type": "string", "unique": True},
        "choreography": {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                },
                "transactions": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "queue": {"type": "string"},
                            "event": {"type": "string"},
                            "order": {"type": ["integer", "string"]},
                            "status": {
                                "type": "string",
                                "enum": [
                                    Status.NOT_STARTED,
                                    Status.IN_PROGRESS,
                                    Status.SUCCESS,
                                    Status.FAILED,
                                    Status.CANCELLED,
                                ],
                            },
                            "message": {"type": "string"},
                            "start": {},
                            "end": {},
                            "runtime": {"type": "number"},
                            "error": {"type": "string"},
                        },
                        "required": ["queue", "event", "order", "status"],
                        "additionalProperties": False,
                    },
                    "additionalProperties": False,
                },
            },
            "required": ["name", "transactions"],
            "additionalProperties": False,
        },
        "status": {
            "type": "string",
            "enum": [
                Status.NOT_STARTED,
                Status.IN_PROGRESS,
                Status.SUCCESS,
                Status.FAILED,
                Status.CANCELLED,
            ],
        },
        "data": {"type": "object"},
        "createdAt": {},
        "updatedAt": {},
    },
    "required": ["sagaId", "choreography", "status", "createdAt", "updatedAt"],
    "additionalProperties": False,
}



"""
testing the saga object
"""
from tc_messageBroker.rabbit_mq.saga.saga_base import get_saga, Saga
from tc_messageBroker.rabbit_mq.saga.choreography_base import IChoreography
from datetime import datetime
from tc_messageBroker.rabbit_mq.saga.transactions import (
    DISCORD_UPDATE_CHANNELS_TRANSACTIONS,
)
from tc_messageBroker.rabbit_mq.status import Status
import numpy as np
from copy import deepcopy
import pytest


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no available MongoDB instance)"
)
def test_inputs():
    connection_url = "mongodb://127.0.0.1:27017/"

    ## we should have this data before running this test in db
    saga = get_saga(
        sagaId="something",
        connection_url=connection_url,
        db_name="Saga",
        collection="saga",
    )

    assert saga.choreography is not None
    assert saga.status in [
        Status.IN_PROGRESS,
        Status.CANCELLED,
        Status.FAILED,
        Status.NOT_STARTED,
        Status.SUCCESS,
    ]
    assert isinstance(saga.created_at, datetime) is True
    assert isinstance(saga.uuid, str) is True


@pytest.mark.skip(
    reason="Unable to test on GitHub Actions (no available MongoDB instance)"
)
def test_saga_update():
    """
    test updating a saga instance in db
    """
    connection_url = "mongodb://127.0.0.1:27017/"

    choreography = IChoreography(
        name="sample", transactions=deepcopy(DISCORD_UPDATE_CHANNELS_TRANSACTIONS)
    )

    saga = Saga(
        choreography=choreography,
        status=Status.NOT_STARTED,
        data={"guildId": "1234"},
        created_at=datetime.now(),
    )

    tx = np.array(choreography.transactions)
    tx[0].status = Status.SUCCESS

    saga._update_save(transactions=tx, mongo_connection=connection_url)



from tc_messageBroker.rabbit_mq.saga.transactions import (
    DISCORD_SCHEDULED_JOB_TRANSACTIONS,
    DISCORD_UPDATE_CHANNELS_TRANSACTIONS,
)
from tc_messageBroker.rabbit_mq.queue import Queue
from tc_messageBroker.rabbit_mq.event import Event
from tc_messageBroker.rabbit_mq.status import Status


def test_discord_update_channels_tx():
    tx = DISCORD_UPDATE_CHANNELS_TRANSACTIONS

    assert len(tx) == 2
    assert tx[0].order == 1
    assert tx[0].queue == Queue.DISCORD_BOT
    assert tx[0].event == Event.DISCORD_BOT.FETCH
    assert tx[0].status == Status.NOT_STARTED

    assert tx[1].order == 2
    assert tx[1].queue == Queue.DISCORD_ANALYZER
    assert tx[1].event == Event.DISCORD_ANALYZER.RUN
    assert tx[1].status == Status.NOT_STARTED


def test_discord_scheduled_job_tx():
    tx = DISCORD_SCHEDULED_JOB_TRANSACTIONS

    assert len(tx) == 1
    # assert tx[0].order == 1
    # assert tx[0].queue == Queue.DISCORD_BOT
    # assert tx[0].event == Event.DISCORD_BOT.FETCH
    # assert tx[0].status == Status.NOT_STARTED

    assert tx[0].order == 1
    assert tx[0].queue == Queue.DISCORD_ANALYZER
    assert tx[0].event == Event.DISCORD_ANALYZER.RUN_ONCE
    assert tx[0].status == Status.NOT_STARTED
